package musr.client.bestfit;

import musr.client.datacolumn.*;
import java.text.DecimalFormat;

/**This is the superclass for all fit classes.
 *Contains methods and variables used by descendant classes.
 **/
public abstract class AbstractFit
{
    /**Takes the values from data and returns them in an array of doubles
     **/
    public static double[] dataToArray(DataColumn data, int start, int end)
    {
        if(start<0)
            start = 0;              //make sure that no value too small is given
        if(end>data.getSize())
            end = data.getSize();    //make sure that no value too large is given
        double[] result = new double[end-start+1];
        //System.out.println("length of result array = " + result.length);
        for(int i=0; i<result.length; i++)
                result[i] = Double.parseDouble(data.getValue(i+start).toString());
        /*if(result==null)
            System.out.println("NULL RESULT BEING RETURNED AT LINE 25");*/
        return result;
    }
    
    public static double[] dataToArray(DataColumn data)
    {
        double[] result = new double[data.getSize()];
        for(int i=0; i<result.length; i++)
            result[i] = Double.parseDouble(data.getValue(i).toString());
        return result;
    }
    
    /**Takes the values from error1 and error2, converts them to double, and returns the sum of their sqares in an array.
     *
     *Throws ErrorBarException when the error bars are null - must deal with this.
     **/
    public static double[] errorSquaredToArray(DataColumn error1, DataColumn error2, int start, int end) throws ErrorBarException
    {   
        //Check to see if error columns are null (used to indicate that no error bars are available)
        if(error1==null||error2==null)
            throw new ErrorBarException("This data has no error bars. " +
                    "\nTherefore, for this fit each point will be equally weighted.");
        double[] result = new double[end-start+1];
        for(int i=0; i<result.length; i++)
        {
            if((i+start)>=error1.getSize())
                throw new ErrorBarException("There are not enough values in the error bars to correspond with the selected range." +
                        "\nTherefore, for this fit each point will be equally weighted.");
            result[i] = Math.pow(Double.parseDouble(error1.getValue(i+start).toString()), 2) 
                + Math.pow(Double.parseDouble(error2.getValue(i+start).toString()), 2);
            if(result[i]==0)    //This will cause division by zero, so best to ignore all the error bars
                throw new ErrorBarException("This fit has one or more error bars with a magnitude of zero, which would give " +
                        "it infinite weight in a best fit. \nTherefore, for this fit each point will be equally weighted.");
        }
        System.out.print("AbstractFit:58   ");
        printArray(result);
        return result;
    }
    
    public static double[] errorSquaredToArray(DataColumn error1, DataColumn error2) throws ErrorBarException
    {
        //Check to see if error columns are null (used to indicate that no error bars are available)
        if(error1==null||error2==null)
            throw new ErrorBarException("This data has no error bars. " +
                    "\nTherefore, for this fit each point will be equally weighted.");
        double[] result = new double[error1.getSize()];
        for(int i=0; i<result.length; i++)
        {
            result[i] = Math.pow(Double.parseDouble(error1.getValue(i).toString()), 2) 
                + Math.pow(Double.parseDouble(error2.getValue(i).toString()), 2);
            if(result[i]==0)    //This will cause division by zero, so best to ignore all the error bars
                throw new ErrorBarException("This fit has one or more error bars with a magnitude of zero, which would give " +
                        "it infinite weight in a best fit. \nTherefore, for this fit each point will be equally weighted.");
        }
        System.out.print("AbstractFit:78   ");
        printArray(result);
        return result;
    }
    
    /**Method used in testing to print 2-D arrays to standard output
     **/
    public static void printMatrix(double[][] matrix)
    {
        if(matrix==null)
            System.out.println("null");
        DecimalFormat formatter = new DecimalFormat("######.0000");
        for(int i=0; i<matrix.length; i++)
        {
            for(int j=0; j<matrix[0].length; j++)
                System.out.print(formatter.format(matrix[i][j]) + "\t");
            System.out.println();
        }
    }
    
    //Method used in testing to print 1-D arrays to standard output
    public static void printArray(double[] array)
    {
        if(array==null)
            System.out.println("null");
        for(int i=0; i<array.length; i++)
        {
            System.out.print(array[i] + "\t");
        }
        System.out.println();
    }
    
    /**Every subclass must have a method to fit the data.
     **/
    public abstract void fit() throws DataNotFittableException;
    
    /**Every subclass must override the default toString() method provided by Object
     **/
    public abstract String toString();
}
