package musr.client.bestfit;

import musr.client.datacolumn.*;

public class SplineFit2 extends AbstractFit
{
    public double[] xData;          //holds input x values
    public double[] yData;          //holds input y values at input x values
    public double[] ypp;            //holds second derivative of spline at input x values
    
    private int n;                  //used in sorting
    
    private String xName;           //used in toString() method
    private String yName;
    
    private String xLabel;
    private String yLabel;
    private String xUnits;
    private String yUnits;

    public static void main(String[] args)
    {
        ConcreteDataColumn testX = new ConcreteDataColumn();
        ConcreteDataColumn testY = new ConcreteDataColumn();
        for(int i=0; i<10; i++)
        {
            testX.addElement(new Double(i));
            testY.addElement(new Double(test(i)));
        }
        SplineFit2 test = new SplineFit2(testX, testY);
        test.fit();
        for(int i=0; i<33; i++)
            System.out.println(test(i/4.0) + "\t" + test.evaluateSplineAt(i/4.0));
    }
    
    public SplineFit2(DataColumn xColumn, DataColumn yColumn)
    {
        xName = xColumn.getName();
        yName = yColumn.getName();
        
        xLabel = xColumn.getLabel();
        yLabel = yColumn.getLabel();
        
        xUnits = xColumn.getUnits();
        yUnits = yColumn.getUnits();
        
        xData = dataToArray(xColumn);
        yData = dataToArray(yColumn);
        heapSort();
    }
    
    private void heapSort()
    {
        n = xData.length;
        buildHeap();
        while(n>1)
        {
            n--;
            swap(0, n);
            downHeap(0);
        } 
    }
    
    private void buildHeap()
    {
        for(int j = n/2-1; j>=0; j--)
            downHeap(j);
    }
    
    private void downHeap(int topIndex)
    {
        int childIndex=2*topIndex+1;
        while (childIndex<n)
        {
            if((childIndex+1<n)&&(xData[childIndex+1]>xData[childIndex]))
                childIndex++;

            if(xData[topIndex]>=xData[childIndex]) return;

            swap(topIndex, childIndex);
            topIndex = childIndex;  
            childIndex = 2*topIndex+1;
        }
    }
    
    private void swap(int i, int j)
    {
        double tempX = xData[i];
        double tempY = yData[i];
        
        xData[i] = xData[j];
        yData[i] = yData[j];
        
        xData[j] = tempX;
        yData[j] = tempY;
    }
    
    public void fit()
    {
        ypp = new double[yData.length];                 //y2
        double[] temp = new double[ypp.length-1];       //u
        double tempSig;                                 //sig
        double tempP;                                   //p
        //un=qn=0 for these splines
        ypp[0] = 0.0;
        temp[0] = 0.0;
        //decomposition - this is a tridiagonal system and therefore wonderfully easy to solve in O(N) operations
        for(int i=1; i<ypp.length-1; i++)
        {
            tempSig = (xData[i]-xData[i-1])/(xData[i+1]-xData[i-1]);
            tempP = tempSig*ypp[i-1]+2.0;
            ypp[i] = (tempSig-1.0)/tempP;
            temp[i] = (yData[i+1]-yData[i])/(xData[i+1]-xData[i]) - (yData[i]-yData[i-1])/(xData[i]-xData[i-1]);
            temp[i] = (6.0*temp[i]/(xData[i+1]-xData[i-1]) - tempSig*temp[i-1])/tempP;
        }
        //backsubstitution - the second part of the algorithm to solve this tridiagonal system
        ypp[ypp.length-1] = 0.0;
        for(int i=ypp.length-2; i>=0; i--)
            ypp[i] = ypp[i]*ypp[i+1] + temp[i];
    }
    
    public String toString()
    {
        return "Spline interpolation of " + yName + " vs " + xName;
    }
    
    /**
     *Evaluates the spline at the given x value.
     */
    public double evaluateSplineAt(double x)
    {
        //make sure that the given x is wiithin the range of the spline - otherwise just return zero
        if((x>=xData[0])&&(x<=xData[xData.length-1]))
        {
            int j = findIndex(x, 0, xData.length-1);
            return a(x,j)*yData[j] + b(x,j)*yData[j+1] + c(x,j)*ypp[j] + d(x,j)*ypp[j+1];
        }
        else
            return 0.0;
    }
    
    /**
     *Finds the index of the first value in xData below the given x value.
     *Since the data is already sorted, this is easier than it may sound...
     */
    private int findIndex(double x, int start, int end)
    {
        if(start==end)
            return start;
        int mid = (start+end)/2;
        if(x==xData[mid])
            return mid;
        else if(x<xData[mid])
            return findIndex(x, start, mid);
        else
            return findIndex(x, mid+1, end);
    }
    
    /**
     *The following functions are used to generate the four components of the cubic spline (namely A, B, C, and D).
     *Each takes two arguments: x is the value at which the function is being evaluated, and j is the index of the spline 
     *at which the function is being evaluated.
     */
    private double a(double x, int j)
    {
        return (xData[j+1]-x)/(xData[j+1]-xData[j]);
    }
    
    private double b(double x, int j)
    {
        return (x-xData[j])/(xData[j+1]-xData[j]);
    }    
    
    private double c(double x, int j)
    {
        double a = a(x, j);
        return (Math.pow(a, 3)-a)*Math.pow(xData[j+1]-xData[j], 2)/6;
    }   
    
    private double d(double x, int j)
    {
        double b = b(x, j);
        return (Math.pow(b, 3)-b)*Math.pow(xData[j+1]-xData[j], 2)/6;
    }
    
    //accessor methods for data column construction
    public String getXLabel()
    {
        return xLabel;
    }
    
    public String getYLabel()
    {
        return yLabel;
    }
    
    public String getXUnits()
    {
        return xUnits;
    }
    
    public String getYUnits()
    {
        return yUnits;
    }
    
    //function used only in testing to determine spline's effectiveness
    private static double test(double x)
    {
        return Math.log10(x*x+1);
    }
}
