package musr.client.csv;
import java.util.*;

public class CSVTokenizer
{
//This class takes in a string representing a row of data in CSV format, parses 
//it into separate entries, and puts these into a vector. These can then be
//returned, either one by one or as the whole vector.
    private String input;       //String being tokenized
    private Vector tokens = null;      //To store tokens
    private int tokensReturned;     //Number of tokens returned
    private char separator = ',';     //Either ',' or ';' depending on options specified when file was exported into .csv format
        
    /**
     Creates a new instance of CSVTokenizer using the given string. This method includes a call to tokenize().
     @param inputString The string to be divided into csv-format tokens.
     @see #tokenize()
     */
    public CSVTokenizer(String inputString)
    {
        tokens = new Vector();
        tokensReturned = 0;        
        
        if(inputString!=null)
        {
            input = inputString;
            if(inputString.trim().length()>0)
                separator = inputString.trim().charAt(inputString.trim().length()-1); //last character of line should be a separator
            this.tokenize();
        }
    }
    /**
     Divides a String (specified in the constructor) into csv-formatted tokens.
     */
    public void tokenize()
    {
        boolean inQuotes = false; //tracks whether iterator is inside quotation marks to check if " is separator of punctuation within entry
        int beginning = 0; //tracks beginning of current token
        for(int i=0; i<input.length(); i++)
        {
            if(input.charAt(i)=='"')        //Token is starting or ending
                inQuotes = !inQuotes;
            else if((input.charAt(i)==separator)&&!(inQuotes))
            {        
                tokens.add((input.substring(beginning, i)).trim());
                beginning = i+1;
            }
        }
        this.removePunctuation();        //Strips the quotes and commas/semicolons off of every token
    }
    
    private void removePunctuation()
    {
        String temp;
        
        for(int i=0; i<tokens.size(); i++)      //Strips the commas (or semicolons) off of every token
        {
            temp = (String)tokens.elementAt(i);
            while(temp.endsWith(Character.toString(separator)))  
                temp = temp.substring(0, temp.length()-1);
            tokens.setElementAt(temp, i);
        }
        
        for(int i=0; i<tokens.size(); i++)      //Strips the trailing and leading quotes off of every token
        {
            temp = (String)tokens.elementAt(i);
            if(temp.length()!=0)
            {
                if(temp.charAt(0)=='"')
                    temp = temp.substring(1);
                if(temp.charAt(temp.length()-1)=='"')
                    temp = temp.substring(0, temp.length()-1);
                tokens.setElementAt(temp,i);
            }
        }
        
        for(int i=0; i<tokens.size(); i++)      //Replaces "" with " within tokens
        {
            temp = (String)tokens.elementAt(i);
            for(int j=0; j<temp.length(); j++)
                if((temp.charAt(j)=='"')&&(temp.charAt(j+1)=='"'))
                    temp = temp.substring(0,j) + temp.substring(j+1);
            tokens.setElementAt(temp, i);
        }
    }
    
    /**
     @return The next token in the tokenizer.
     */
    public String getNextToken()
    {
        if((this.tokens.size()>0)&&(this.hasMoreTokens()))
        {
            String result = (String)tokens.elementAt(tokensReturned);
            tokensReturned++;
            if((result==null)||(result.equals("")))
                result = " ";
            return result;
        }
        else
            return null;
    }
    
    /**
     @return True if there are more tokens in the tokenizer, false otherwise.
     */
    public boolean hasMoreTokens()
    {
        if(this.tokens!=null)
            return (this.tokensReturned<this.getTokenCount());
        else
            return false;
    }
    /**
     @return The initial number of tokens in the tokenizer
     */
    public int getTokenCount()
    {
        if(this.tokens!=null)
            return this.tokens.size();
        else
            return -1;
    }
    
    /**
     @return A string representation of all the tokens initially present in the tokenizer.
     */
    public String toString()
    {
        if(this.tokens!=null)
        {
            String result = "";
            for(int i=0; i<tokens.size(); i++)
                result = result + (String)tokens.elementAt(i);
            return result;
        }
        else
            return null;
    }
    /**
     @return A Vector holding all the tokens initially present in the tokenizer.
     */
    public Vector getTokenVector()
    {
        return (Vector)this.tokens.clone();
    }
    /**
     @return The number of tokens already returned.
     */
    public int getTokensReturned()
    {
        return this.tokensReturned;
    }
    /**
     @return The original input String.
     */
    public String getInput()
    {
        return this.input;
    }
}