package musr.client.fourier;

/**
 *This class holds complex numbers in the form of two doubles - one stores the real component, the other the imaginary.
 *It also contains the methods needed to perform basic arithmetic and other simple mathematical operations on complex numbers.
 */
public class Complex implements Cloneable
{
    private double real;
    private double imaginary;
    
    /**
     *Default constructor; initializes an Complex with a numerical value of zero.
     */
    public Complex()
    {
        real=0;
        imaginary=0;
    }
    
    /**
     *Initializes a Complex with the parameters as the real and imaginary components.
     */
    public Complex(double newReal, double newImaginary)
    {
        real = newReal;
        imaginary = newImaginary;
    }
    
    /**
     *Initializes a Complex with the parameters as the real and imaginary components.
     */
    public Complex(Double newReal, Double newImaginary)
    {
        real = newReal.doubleValue();
        imaginary = newImaginary.doubleValue();
    }
    
    /**
     *Returns a string representation of the complex number.
     *
     *The value returned by this method has the form "[real component] + [imaginary component]i".
     */
    public String toString()
    {
        return real + "+" + imaginary + "i";
    }
    
    /**
     *Returns true if otherObject represents a Complex object numerically equal to this one; otherwise it returns false.
     */
    public boolean equals(Object otherObject)
    {
        if(this==null)
            return (otherObject==null);
        else if(otherObject==null)      //this!=null, or otherwise the first condition (this==null) would have been true
            return false;
        else
            return((this.real == ((Complex)otherObject).real) && (this.imaginary == ((Complex)otherObject).imaginary));
    }
    
    /**
     *Returns an Object (also of type Complex) that is numerically equal to the calling object.
     */
    public Object clone()
    {
        return new Complex(this.real, this.imaginary);
    }
    
    /**
     *Returns the real component.
     */
    public double realPart()
    {
        return this.real;
    }
    
    /**
     *Returns the imaginary component.
     */
    public double imaginaryPart()
    {
        return this.imaginary;
    }
    
    /**
     *Returns the magnitude, or real modulus.
     */
    public double magnitude()
    {
        return this.real*this.real + this.imaginary*this.imaginary;
    }
    
    /**
     *Returns the complex conjugate.
     */
    public Complex conjugate()
    {
        return new Complex(this.real, -1*this.imaginary);
    }
    
    /**
     *Returns the Complex exponential of the given value.
     */
    public Complex exp()
    {
        return new Complex(Math.exp(this.real)*Math.cos(this.imaginary), Math.exp(this.real)*Math.sin(this.imaginary));
    }
    
    /**
     *Returns the Complex representing the numerical sum of this Complex and the Complex given as the parameter.
     */
    public Complex add(Complex other)
    {
        return new Complex(this.real+other.real, this.imaginary+other.imaginary);
    }
    
    /**
     *Adds the value of the Complex other to this Complex.
     */
    public void addTo(Complex other)
    {
        this.real+=other.real;
        this.imaginary+=other.imaginary;
    }
    
    /**
     *Subtracts the value of the Complex other from this Complex.
     */    
    public Complex subtract(Complex other)
    {
        return new Complex(this.real-other.real, this.imaginary-other.imaginary);
    }
    
    /**
     *Multiplies by the given Complex value.
     */
    public Complex multiply(Complex other)
    {
        return new Complex(this.real*other.real-this.imaginary*other.imaginary, this.real*other.imaginary+this.imaginary*other.real);
    }
    
    /**
     *Multiplies by the given real value.
     */
    public Complex multiply(double number)
    {
        return new Complex(this.real*number, this.imaginary*number);
    }
    
    /**
     *Returns the multiplactive inverse as a new Complex.
     */
    public Complex inverse()
    {
        return new Complex(this.real/this.magnitude(), -1*this.imaginary/this.magnitude());
    }
    
    /**
     *Divides by the given Complex value.
     */
    public Complex divide(Complex other)
    {
        return this.multiply(other.inverse());
    }
    
    /**
     *Divides by the given real value.
     */
    public Complex divide(double number)
    {
        return this.multiply(1.0/number);
    }
    
    /**
     *Returns the trigonometric sine. The calling object is assumed to represent a complex angle measured in radians.
     */
    public Complex sine()
    {
        return new Complex(Math.sin(this.real)*Math.cosh(this.imaginary), Math.cos(this.real)*Math.sinh(this.imaginary));
    }
    
    /**
     *Returns the trigonometic cosine. The calling object is assumed to represent a complex angle measured in radians.
     */
    public Complex cosine()
    {
        return new Complex(Math.cos(this.real)*Math.cosh(this.imaginary), -1*Math.sin(this.real)*Math.sinh(this.imaginary));
    }
    
    /**
     *Returns the complex phase.
     */
    public double phase()
    {
        return Math.atan(this.imaginary/this.real);
    }
    
    /**
     *Sets the value of this Complex object to have the specified real and imaginary components.
     */
    public void setValue(double newReal, double newImaginary)
    {
        this.real = newReal;
        this.imaginary = newImaginary;
    }
    
    /**
     *Sets the value of this Complex object to have the real component specified by the first element of the double array and the
     *imaginary component specified by the second parameter.
     *
     *@param newValue  A two-element double array with the new real component as its first element and the new imaginary component
     *                     as its second.
     */
    public void setValue(double[] newValue)
    {
        this.real = newValue[0];
        this.real = newValue[1];
    }
}
