/*Note: This class requires that the input data be complex-valued and have a length that is a power of two. The former requirement is
 *basic enough to understand and simple enough to circumvent (for real-valued data, all imaginary values are zeros) but the latter depends
 *on the particulars of the fast Fourier transform (FFT) algorithm used to process the data quickly. This is why the padding of zeroes is added to
 *the end of each column. For more information, consult a reference text on the FFT in programming, such as 'Numerical Recipes in C++'.*/

package musr.client.fourier;
import javax.swing.*;
import musr.client.datacolumn.*;
import java.util.Vector;
import musr.client.function.Function;
import musr.client.gui.GUI;
import musr.client.dbreader.Constants;
import musr.client.dbreader.Functions;

/**
 *This class uses a fast Fourier transform algorithm to resolve time data discretely sampled at constant intervals into a discrete frequency
 *spectrum. For more on the fast Fourier transform algorithm, consult <i>Numerical Recipes in C++: The Art of Scientifc Computing</i>. 
 */
public class FastFourierTransform
{
    private Complex[] data;    //Holds the input values of the original discrete function - [k][0] real, [k][1] imaginary
    private int log2DataLength; //Stores the base-2 exponent of data.length AFTER having padded it with zeroes to ensure a power-of-2 length
    private int dataLength;     //Stores data.length; prevents having to access the array variable length over and over in the loops
    
    /**
     *Used to pad arrays, and also as a default value in error recovery.
     */
    public Complex ZERO;
    
    //main method for testing purposes only
    
    public static void main(String[] args) throws Exception
    {
        double[][] test = new double[1600][2];
        for(int i=1; i<test.length; i++)
        {
            test[i][0] = Math.cos(400*Math.PI/i);
            test[i][1] = Math.sin(400*Math.PI/i);
            //System.out.println(test[i][0] + " + " + test[i][1] + "i");
        }
        FastFourierTransform t = new FastFourierTransform(test);
        long start = System.currentTimeMillis();
        Complex[] result = t.transform();
        System.out.println("Time to finish: " + (System.currentTimeMillis()-start));
        /*for(int i=0; i<result.length; i++)
            System.out.println(result[i].toString());*/
    }
    
    /**
     *This constructor takes a two-dimensional array of values (representing the real and imaginary components of a complex array of values)
     *and uses it to intialize all the variables needed to perform the Fourier transform.
     */
    public FastFourierTransform(double[][] input)
    {       
        ZERO = new Complex(0, 0);
        int inputLength = input.length;
        //System.out.println("input length: " + inputLength);
        
        if((inputLength>100000)&&(JOptionPane.showConfirmDialog(null, "This data set is very large. \nIt is entirely possible that the Fourier" +
                    " transform will take a very long time to complete. \nOr it could cause your computer to freeze. \n\nAre you sure you want" +
                    " to continue?", "WARNING: Large data set.", JOptionPane.YES_NO_OPTION)==JOptionPane.NO_OPTION))
        {
            //Put in some meaningless garbage data
            data = new Complex[]{ZERO, ZERO};
            dataLength = 2;
            log2DataLength = 1;
        }
        
        else
        {
            log2DataLength = (int)Math.floor(Math.ceil(Math.log(input.length)/Math.log(2)));
            dataLength = (int)Math.pow(2, log2DataLength);
            data = new Complex[dataLength];

            //Fill the array with data, then once there is no more data, with zeroes (to a power-of-2 length):
            for(int i=0; i<inputLength; i++)
                data[i] = new Complex(input[i][0], input[i][1]);
            for(int i=inputLength; i<dataLength; i++)
                data[i] = ZERO;
        }
    }
    
    /**
     *This constructor takes two DataColumns(representing the real and imaginary components of a complex data series) and uses them to 
     *initialize all the variable needed to perform the Fourier transform.
     */
    public FastFourierTransform(DataColumn real, DataColumn imaginary)
    {
        System.out.println("FastFourierTransform constructor called");
        ZERO = new Complex(0,0);
        int inputLength = real.getSize();
        if(inputLength!=imaginary.getSize())
        {
            JOptionPane.showMessageDialog(null, "The DataColumns containing the real and imaginary data are not of equal length.\n" +
                    "It is impossible to form a set of complex data if there are not equal numbers of real and imaginary components\n." +
                    "Please try again, selecting the correct real and imaginary comlumns.");
            //Put in some meaningless garbage data
            data = new Complex[]{ZERO, ZERO};
            dataLength = 2;
            log2DataLength = 1;
        }
        
        else
        {
            log2DataLength = (int)Math.floor(Math.ceil(Math.log(real.getSize())/Math.log(2)));
            dataLength = (int)Math.pow(2, log2DataLength);
            data = new Complex[dataLength];
            
            //Fill the array with data, then once there is no more data, with zeroes (to a power-of-2 length):
            for(int i=0; i<inputLength; i++)
                data[i] = new Complex(Double.parseDouble(real.getValue(i).toString()), Double.parseDouble(imaginary.getValue(i).toString()));
            for(int i=inputLength; i<dataLength; i++)
                data[i] = ZERO;
        }
    }
    
    //Reverses the bits in the data array (part of the FFT algorithm)
    private void bitReverse()
    {
        //Placeholder variable for individual swaps of complex pairs
        Complex placeholder = new Complex();
        
        int newLocation;
        
        boolean[] alreadyMoved = new boolean[dataLength];     //Ensures that reverse swaps don't occur
      
        for(int i=0; i<dataLength; i++)
        {
            if(!alreadyMoved[i])
            {
                newLocation = Integer.parseInt(reverse(Integer.toBinaryString(i)),2);
                if(!alreadyMoved[newLocation])
                {
                    alreadyMoved[newLocation] = true;
                    alreadyMoved[i] = true;

                    placeholder = data[i];
                    data[i] = data[newLocation];                
                    data[newLocation] = placeholder;
                    //System.out.println("Index " + i + " moved to " + newLocation);
                }
            }
        }
    }
    
    //Returns the String in reverse order of input (eg, "100111" goes to "111001")
    private String reverse(String input)
    {
        if(log2DataLength==0)       //This would mean that input is either "0" or "1", hopefully...
            return input;
        //System.out.print(input + " =====> ");
        while(input.length()<log2DataLength)
            input = "0" + input;
        //System.out.print(input + " =====> ");
        String result = input.substring(log2DataLength-1);
        for(int i=log2DataLength-2; i>=0; i--)
            result = result + input.charAt(i);
        //System.out.println(result);
        return result;
    }
    
    /**
     *Performs the Fourier transform on the complex time data input. The algorithm used for the fast Fourier transform is especially ingenious; 
     *instead of approximating the integral (which is an O(N^2) process) it subdivides the interval repeatedly, turning the transform into an
     *O(N log2(N)) process.
     *
     *@return An array of complex numbers represnting the output of the Fourier transform.
     */
    public Complex[] transform()
    {
        //order the bits of the indices in reverse order
        //long start = System.currentTimeMillis();
        this.bitReverse();
        //System.out.println("Time for bit reversal: " + (System.currentTimeMillis()-start));
        //The new array holds ordered pairs of complexes with k as the first index
        Complex[] result = new Complex[dataLength]; 
        //Fill up the array:
        //start = System.currentTimeMillis();
        Complex wValue = new Complex(Math.cos(2*Math.PI/dataLength), Math.sin(2*Math.PI/dataLength));
        Complex wToK = new Complex(1, 0);
        //long loopStart = System.currentTimeMillis();
        //long  multiplyStart;
        for(int k=0; k<dataLength; k++)
        {
            //System.out.println("k = " + k + ", ");
            result[k] = (Complex)(fourier(wToK).clone());
            //System.out.println("Fk = " + result[k]);
            //System.out.println();
            //System.out.println();
            wToK = wToK.multiply(wValue);
        }
        /*for(int i=0; i<result.length; i++)
            System.out.println("F" + i + " = " + result[i]);*/
        //System.out.println("Time to complete array loop: " + (System.currentTimeMillis()-loopStart));
        //System.out.println("Total time to fill result array: " + (System.currentTimeMillis()-start));
        return result;
    }
    
    private Complex fourier(Complex wToK)
    {
        Complex[] f = new Complex[dataLength];
        for(int i=0; i<dataLength; i++)
            f[i] = data[i];
        //System.out.println(data[0] + "--" + f[0]);
        int increment; 
        for(int j=1; j<log2DataLength; j++)
        {
            increment = (int)Math.pow(2, j);
            for(int c=0; c<dataLength; c+=increment)
                f[c].addTo((f[c+increment/2]).multiply(wToK));
        }
        //System.out.println("returning " + f[0]);
        return f[0];
    }
    
    /**
     *Accessor method for the number of data elements in the data to be Fourier transformed.
     */
    public int getDataLength()
    {
        return dataLength;
    }
    
    /**
     *Applies the given window function to the data before performing the Fourier transform.
     */
    public void applyWindowFunction(String input, DataColumn timeColumn, GUI parentGUI, Functions functionsTable, Constants constantsTable)
    {
        System.out.println("Applying apodization");
        if(input.length()>0)
        {
            Complex[] result = new Complex[dataLength];
            Function apodizationFunction = new Function(input);
            
            if(timeColumn==null)        //use the bin numbers to evaluate the apodization function
            {
                for(int i=0; i<data.length; i++)
                    result[i] = data[i].multiply(apodizationFunction.evaluate(parentGUI, (double)i, false, functionsTable, constantsTable));
            }
            else                        //use the values in the time column to evaluate the apodization function
            {
                for(int i=0; i<data.length; i++)
                    result[i] = data[i].multiply(apodizationFunction.evaluate(parentGUI, 
                            Double.parseDouble(timeColumn.getValue(i).toString()), false, functionsTable, constantsTable));
            }
            
            data = result;
        }
    }
}