package musr.client.plot2d;
import java.applet.*;
import java.awt.*;

/**
 * Contains all the necessary information to be able to draw an axis. Such as, how many tics it will have,
 * it's orientation, graph settings information and so on.
 * The actual axis is drawn in the subclasses that extend Axis.
 * min, max, numTics, pixMin, pixMax - same properties as the defined variables
 * boolean typeAxis   - true if LogAxis, false if LinearAxis or InverseAxis
 * optionVariables - holds various variables for plotting
 *
 * @author Ivan Tucakov, Nikola Tucakov
 */
public abstract class Axis extends Applet implements java.io.Serializable {

  static final long serialVersionUID = 0L; // SUID
	/* Instance Variables */
        /**
         * Minimal and maximal value of data
         */
	protected double min, max;
        /**
         * Number of tics
         */
	protected int numTics;
        /**
         * Array of tics for one Axis.
         * They get loaded in Axis constructor. LinearAxis uses setTicsLooseLabel() method and LogAxis constructs it on the spot. They both load 'nice' values
         */
	protected Tic[] tics;
        /**
         * distance between two tics
         */
	protected double ticSpacing;

        /**
         * Number of pixels in a unit value (for lin)
         * Number of pixels between 10^x (for log)
         */
	protected double scale;
        /**
         * Starting and ending point of the "Plot" in pixels
         */
	protected int pixMin, pixMax;
        /**
         * Distance from the edge of the "Window" in pixels
         */
	protected int pixDist;
        /**
         * Orientation of the Axis
         */
	protected String orient;
        /**
         * Number of fractional digits in labels
         */
	protected int numFraction;

        /**
         * What to draw as a Tic (dash, dot)
         */
	protected String ticOutput;
        /**
         * Whether to output the label
         */
	protected boolean labelOutput;
        /**
         * What side of the Axis to draw the Tics
         * (down, up, left, right)
         */
	protected String ticSide;

        /**
         * Height of the Applet
         */
	protected int appletHeight;
        /**
         * Width of the Applet
         */
	protected int appletWidth;

        /**
         * Holds various Option Variables
         */
	protected PlotOptionsFrame optionVariables;

        /**
         * This is used to set graphics stroke (setStroke()), i.e. make a dashed line
         */
        private class BasicStrokeA extends java.awt.BasicStroke implements java.io.Serializable {
          public BasicStrokeA(float width, int cap, int join, float miterlimit, float[] dash, float dash_phase) {
            super(width, cap, join, miterlimit, dash, dash_phase);
          }
        };
        BasicStrokeA dashed;

	/* Constructor methods */
	public Axis(double min, double max, int numTics, int pixMin, int pixMax, boolean typeAxis, PlotOptionsFrame optionVariables) {
		// initialize instance variables
		this.min = min; this.max = max;
		this.numTics = numTics;
		this.pixMin = pixMin; this.pixMax = pixMax;
		this.optionVariables = optionVariables;

		numFraction=-1;
		// If it is LinearAxis or InverseAxis, use the setTicsLooseLabel function
		if (!typeAxis) setTicsLooseLabel();
		else {
                                              // If it is Logarithmic, load the integer values into the tics array
                                              // Note: the values are integer because we are taking Math.floor value of the double (i.e. min )
                                              tics = new Tic[numTics];
		  for(int i=0; i<numTics;  i++)  tics[i] = new Tic(Math.floor(min+i));
//System.out.println("Axis-> i: "+i+" min: "+min+" math.floor: "+Math.floor(min+i));}
		   ticSpacing = 1;
		}
                setScale(pixMax-pixMin);
		pixDist = 0;
		orient = new String("horizontal");
		// In case not initialised
		appletHeight =  appletWidth = 500;

		ticOutput = new String("dash");
		labelOutput = true;
		ticSide = new String("up");

                // This is used to set graphics stroke (setStroke()), i.e. make a dashed line
                float dash1[] = {4.0f}; // this is the length and the spacing between dashed lines
                dashed = new BasicStrokeA(1.0f,
                                                     BasicStroke.CAP_BUTT,
                                                     BasicStroke.JOIN_MITER,
                                                     10.0f, dash1, 0.0f);
	}
	public Axis() {
			this(0.0, 0.0, 0, 100, 500, false, new PlotOptionsFrame());
	}


	/* Instance Methods */

        /**
         * abstract, defined in it's subclasses
         */
	public abstract void drawAxis(Graphics g);


	/**
         *  This method computes tics using loose labels method.
         *  used for linear or inverse axis, loads the Tics array with appropriate values
         */
	public void setTicsLooseLabel() {
	 try {
    // Local Variables
		double Graphmin, Graphmax;
		double Range;

		// Set Parameters for computation
		Range = niceNum(Math.abs(max-min), false);
		ticSpacing = niceNum(Range/(numTics - 1), true);
		Graphmin = (double) Math.floor(min/ticSpacing)*ticSpacing;
		Graphmax = (double) Math.ceil(max/ticSpacing)*ticSpacing;
		if (numFraction==-1)
			numFraction = (int) Math.max(-Math.floor(Tools.log10(ticSpacing)), 0.0);

		// Count the number of New tics
		double x;
    for(x=Graphmin, numTics=0; x<Graphmax+0.5*ticSpacing;
		x+=ticSpacing, numTics++);

    // Load the data into the tics array
    tics = new Tic[numTics];
    int i;
    for(x=Graphmin, i=0; x<Graphmax+0.5*ticSpacing; x+=ticSpacing, i++) {
		if (i<numTics+1) tics[i] = new Tic(x);
//System.out.print(x+",");
    }
//System.out.print("\n");
	 }
		catch (Exception e) {System.err.println(e.getMessage());}
	}

	/**
         * A helper function for loose_lable function
         * This method finds a "nice" number approximately equal to x.
         * Round the number if round = true, take ceiling if round = false.
         */
	public double niceNum(double x, boolean round) throws Exception {
		int exp;  // Exponent of x
		double f;  // Fractional part of x
		double nf;  // Nice, rounded fraction

		exp = (int) Math.floor(Tools.log10(x));
		f = x/Math.pow(10.0, exp);

		if (round) {
			if (f<1.5) nf = 1.0;
      else if (f<3.0) nf = 2.0;
			else if (f<7.0) nf = 5.0;
			else nf = 10.0;
		}
		else {
      if (f<=1.0) nf = 1.0;
      else if (f<=2) nf = 2.0;
      else if (f<=5) nf = 5.0;
      else nf = 10.0;
		}
		return nf*Math.pow(10.0, exp);
	}

	/**
         * This method returns the number of pixels that corresponds to given value
         *
         * Major BUG!!! Fixed:
         *    The value pixMin along x axis is the left distance. BUT!! along y axis, it is actually
         *    the top distance, not bottom distance! This value used to be added to the overall pix value.
         *    That's a mistake since the y value needs the bottom pix distance, not the top one
         *    THerefore, I added the boolean variable that
         *    is passes to the method to determine wether the axis is X or Y so that the minimum value
         *    would be added to the tic value accordingly. i.e. either optionVariables.leftDistance
         *     for X axis or optionVariables.downDistance for Y axis (instead of pixMin for both).
         */
        public int pix(double value, boolean xAxis) {
          // If the Limits are strict
          // The starting numeric value in regards to which everything else is calculated
          // is the min value of all the data to be plotted:
          if (optionVariables.strictLimitsMenuItem.getState()) {
            if (xAxis) return (int) ( (value - min) * scale + optionVariables.leftDistance ); //it is x axis
            else return (int) ( (value - min) * scale + optionVariables.downDistance ); // it is y axis
          // If the limits are not strict
          // The starting numeric value in regards to which everything else is calculated
          // is the first value of the tics array: tics[0].getValue
          } else {
            if (xAxis) return (int) ( (value - tics[0].getValue()) * scale + optionVariables.leftDistance );
            else return (int) ( (value - tics[0].getValue()) * scale + optionVariables.downDistance );
          }                        // (value - min_tic)
        }

	/**
         * This method returns value that corresponds to given pixels
         * note: add boolean value xAxis so that the method would know whether it is
         * vertical or horizontal axis
         */
	public double value(int Pix, boolean xAxis) {
		try {                                 // scale should be initialized!
			//      (Pix - pixMin)/scale + tics[0].getValue();
		}                          // + min_tic
		catch (ArithmeticException e) {
			System.out.println(e.getMessage() + ", scale=0");
			System.exit(-1);
		}
//System.out.println("axis value: "+(Pix - pixMin)/scale + min+" min axis value: "+min);
                // take the min limiting value
                if (optionVariables.strictLimitsMenuItem.getState()) {
                  if (xAxis) return (Pix - optionVariables.leftDistance)/scale + min;
                  else return (Pix - optionVariables.downDistance)/scale + min;
                // take the first tic[] value
                } else {
                  if (xAxis) return (Pix - optionVariables.leftDistance)/scale + tics[0].getValue();
                  else return (Pix - optionVariables.downDistance)/scale + tics[0].getValue();
                }
	}                              // + min_tic





	// Methods that access Instance Variables
	// Declare final not to get dynamic method lookup,
	//        and could be initialised by the compiler
	public final double getMin() { return min; }
	public final void setMin(double d) { min = d; }

	public final double getMax() { return max; }
	public final void setMax(double d) { max = d; }

	public final int getNumTics() { return numTics; }
	public final void setNumTics(int i) { numTics = i; }

	public final Tic getTic(int i) { return tics[i]; }

	public final double getScale() { return scale; }

        /**
         * Sets the pixel value of the values between two Tics
         */
        public final void setScale(int PixRange) {
          // Here, it is calculated with respect to the min and max data value to be plotted
          if (optionVariables.strictLimitsMenuItem.getState())
            scale = (double) PixRange /(max - min);
          // Here, it is calculated with respect to the first and the last tic value
          else
            scale = (double) PixRange /
		(tics[numTics-1].getValue() - tics[0].getValue());
	} //(max_tic                          - min_tic )

	public final int getPixMin() { return pixMin; }
	public final void setPixMin(int i) { pixMin = i; setScale(pixMax-pixMin);}

	/**
         * This method returns pixels that corresponds to given value
         * note: add boolean value xAxis so that the method would know whether it is
         * vertical or horizontal axis
         */
	public int pixel(double value, boolean xAxis) {

                // take the min limiting value
                if (optionVariables.strictLimitsMenuItem.getState()) {
                  if (xAxis) return Math.round( ((float)value - (float)min) *
                                                            (float)scale +
                                                            Float.parseFloat(Long.toString(optionVariables.leftDistance)));
                  else return Math.round( ((float)value - (float)min) *
                                                            (float)scale +
                                                            Float.parseFloat(Long.toString(optionVariables.downDistance)));
                // take the first tic[] value
                } else {
                  if (xAxis) return Math.round( ((float)value - (float)tics[0].getValue()) *
                                                            (float)scale +
                                                             Float.parseFloat(Long.toString(optionVariables.leftDistance)));
                  else return Math.round( ((float)value - (float)tics[0].getValue()) *
                                                            (float)scale + Float.parseFloat(Long.toString(optionVariables.downDistance)));
                }
            }

	public final int getPixMax() { return pixMax; }
	public final void setPixMax(int i) { pixMax = i; setScale(pixMax-pixMin);}


	public final void setPixDist(int i) { pixDist = i; }

	public final void setOrient(String s) { orient = s; }

	public final void setNumFraction(int i) { numFraction = i; }

	public final void setTicOutput(String s) { ticOutput = s; }

	public final void setLabelOutput(boolean b) { labelOutput = b; }

	public final void setTicSide(String s) { ticSide = s; }

	public final void setAppletWidthAndHeight(Dimension D) {
		appletHeight = D.height;
      appletWidth = D.width;
	}

}


