package musr.client.spline;

import java.util.Date;
import musr.client.datacolumn.*;

/**
 *This class is used to create B-splines, which are interpolations of the given data set using linear combinations of four cubic basis 
 *polynomials. The input and output are both in the form of DataColumns.
 */
public class BSpline
{
    //holds the x and y data from the original data columns
    private Point[] points;
    
    //the number of interpolated points on each interval
    private int steps;
    
    ///used to hold the result during the getSplines() method - global so that they can be initialized using the name, label, and units
    //of the original data columns
    private ConcreteDataColumn xColumn;
    private ConcreteDataColumn yColumn;
    
    //used in heapsort
    private int n; 
    
    //parameters set by the user to control the shape of the spline
    private int multiplicity;
    private double weight;
    
    //main method, for testing
    
    /*public static void main(String[] args)
    {
        ConcreteDataColumn testX = new ConcreteDataColumn();
        ConcreteDataColumn testY = new ConcreteDataColumn();
        for(int i=-10; i<10; i++)
        {
            testX.addElement(new Double(i));
            testY.addElement(new Double(f(i)));
        }
        DataColumn[] result = new BSpline(testX, testY).getSplines();
        for(int i=0; i<result[0].getSize(); i++)
            System.out.println(result[0].getValue(i)+"\t"+result[1].getValue(i));
    }*/
    
    //test function
    private static double f(double x)
    {
        if(x!=0.0)
            return 1.0/Math.sin(x);
        else
            return x;
    }
    
    /**
     *This is a constructor for a new BSpline. It initializes all the variables required to produce the spline columns.
     *@param xData This is the x-column used by the spline.
     *@param yData This is the y-column used by the spline.
     */
    public BSpline(DataColumn xData, DataColumn yData, int newMultiplicity, double newWeight)
    {
        multiplicity = newMultiplicity;
        weight = newWeight;
        if(weight==0)
            weight = Double.MIN_VALUE;      //prevent div-by-zero errors
        
        points = new Point[xData.getSize()];
        for(int i=0; i<points.length; i++)
            points[i] = new Point(Double.parseDouble(xData.getValue(i).toString())*weight, 
                    Double.parseDouble(yData.getValue(i).toString())*weight);
        
        //sort the array by ascending x values so that the splines connect the points in order
        n = points.length;
        heapSort();
        
        //The endpoints need a "multiplicity" (i.e. a repetition) equal to the degree of the polynomial - so do that:
        Point[] temp = new Point[points.length*multiplicity+4];
        for(int i=0; i<points.length; i++)
            for(int j=0; j<multiplicity; j++)
                temp[multiplicity*i+j+2] = points[i];
        temp[0] = points[0];
        temp[1] = points[0];
        temp[temp.length-2] = points[points.length-1];
        temp[temp.length-1] = points[points.length-1];
        points = temp;
        
        xColumn = new ConcreteDataColumn(xData.getName(), xData.getLabel(), xData.getUnits(), new Date());
        yColumn = new ConcreteDataColumn(yData.getName(), yData.getLabel(), yData.getUnits(), new Date());
        
        //make the number of interpolative steps on each interval such that there are approximately 255 steps in total
        steps = 255/points.length;
    }
    
    //returns basis functions for B-spline polynomial, evaluated at t
    private double bFunction(int i, double t)
    {
        switch(i)
        {
            case -2:
                return (((-t+3)*t-3)*t+1)/6;
            case -1:
                return (((3*t-6)*t)*t+4)/6;
            case 0:
                return (((-3*t+3)*t+3)*t+1)/6;
            case 1:
                return (t*t*t)/6;
            default:
                return 0;
        }
    }
    
    //evaluates the given spline
    private Point evalPoint(int i, double t)
    {
        double x = 0;
        double y = 0;
        
        for(int j=-2; j<=1; j++)
        {
            x += bFunction(j, t)*points[i+j].x;
            y += bFunction(j, t)*points[i+j].y;
        }
        
        return new Point(x, y);
    }
    
    /**
     *Makes the spline columns and returns them in a two-element array of DataColumns.
     *The first element of this array represents the x values of the spline columns, while the second represents the y values.
     *Both are in the form of MultiColumns with zero error bars all the way through.
     *
     *@return A two-element array of DataColumns, holding the x and y columns (both of type MultiColumn) in the first and second positions
     *          of the array, respectively.
     */
    public DataColumn[] getSplines()
    {
        //in the finished product, the columns will both be MultiColumns
        //these require error bars, so ,make some... 
        ConcreteDataColumn zeroColumn = new ConcreteDataColumn();
        Point temp;
        for(int i=2; i<points.length-1; i++)
        {            
            for(int j=1; j<=steps; j++)
            {
                temp = evalPoint(i, j/(double)steps);
                xColumn.addElement(new Double(temp.x/weight));
                yColumn.addElement(new Double(temp.y/weight));
            }
        }
        Double zero = new Double(0.0);
        for(int i=0; i<xColumn.getSize(); i++)
            zeroColumn.addElement(new Double(zero));
        return new DataColumn[]{new MultiColumn(xColumn, zeroColumn, zeroColumn), new MultiColumn(yColumn, zeroColumn, zeroColumn)};
    }
    
        /**
     *This method sorts data using a heap sort algorithm.
     *This was used instead of faster algorithms such as quick sort because quick sort is slowest with already-sorted data, and there is a
     *nontrivial chance that the x data is already sorted.
     */
    private void heapSort()
    {
        buildHeap();
        while(n>1)
        {
            n--;
            swap(0, n);
            downHeap(0);
        } 
    }

    private void buildHeap()
    {
        for(int j = n/2-1; j>=0; j--)
            downHeap(j);
    }

    private void downHeap(int topIndex)
    {
        int childIndex=2*topIndex+1;
        while (childIndex<n)
        {
            if((childIndex+1<n)&&(points[childIndex+1].x>points[childIndex].x))
                childIndex++;

            if(points[topIndex].x>=points[childIndex].x) return;

            swap(topIndex, childIndex);
            topIndex = childIndex;  
            childIndex = 2*topIndex+1;
        }
    }

    private void swap(int i, int j)
    {
        Point temp = points[i];
        points[i] = points[j];
        points[j] = temp;
    }
}